#include "include/ioman.h"
#include <kernel.h>
#include <string.h>
#include <malloc.h>
#include <stdio.h>

#define MAX_IO_REQUESTS 64
#define MAX_IO_HANDLERS 64

extern void * _gp;

static int gIOTerminate = 0;

#define THREAD_STACK_SIZE   (8 * 1024) 

static u8          disp_stack[THREAD_STACK_SIZE] ALIGNED(16); 
static u8          thread_stack[THREAD_STACK_SIZE] ALIGNED(16); 

struct io_request_t {
	int  type;
	void *data;
	struct io_request_t *next;
};

struct io_handler_t {
	int  type;
	io_request_handler_t handler;
};

/// Circular request queue
static struct io_request_t *gReqList;
static struct io_request_t *gReqEnd;

static struct io_handler_t gRequestHandlers[MAX_IO_HANDLERS];


static void ioThreadDispatcher ( void* apParam );
static void ioAlarmFunc(s32 id, u16 time, void *arg);

static int gHandlerCount;

// id of the processing thread
static s32 gThreadId;
// id of the thread that handles thread switching
static s32 gDispatcherThreadID;

// lock for tip processing
static s32 gProcSemaId;
// lock for queue end
static s32 gEndSemaId;
// ioPrintf sema id
static s32 gIOPrintfSemaId;

static ee_thread_t gThread;
static ee_sema_t gQueueSema;

int ioRegisterHandler(int type, io_request_handler_t handler) {
	WaitSema(gProcSemaId);
	
	if (handler == NULL)
		return IO_ERR_INVALID_HANDLER;
	
	if (gHandlerCount >= MAX_IO_HANDLERS)
		return IO_ERR_TOO_MANY_HANDLERS;

	int i;
	
	for (i = 0; i < gHandlerCount; ++i) {
		if (gRequestHandlers[i].type == type)
			return IO_ERR_DUPLICIT_HANDLER;
	}
	
	gRequestHandlers[gHandlerCount].type = type;
	gRequestHandlers[gHandlerCount].handler = handler;
	gHandlerCount++;
	
	SignalSema(gProcSemaId);
	
	return IO_OK;
}

static io_request_handler_t ioGetHandler(int type) {
	int i;

	for (i = 0; i < gHandlerCount; ++i) {
		struct io_handler_t* h = &gRequestHandlers[i];
		
		if (h->type == type)
			return h->handler;
	}
	
	return NULL;
}

static void ioProcessRequest(struct io_request_t* req) {
	if (!req)
		return;
	
	io_request_handler_t hlr = ioGetHandler(req->type);
	
	// invalidate the request
	void *data = req->data;

	if (hlr)
		hlr(data);
}

static void ioWorkerThread(void) {
	while (!gIOTerminate) {
		SleepThread();

		// do we have a request in the queue?
		while (gReqList) {
			WaitSema(gProcSemaId);
			
			struct io_request_t* req = gReqList;
			ioProcessRequest(req);

			// lock the queue tip as well now
			WaitSema(gEndSemaId);
			
			// can't be sure if the request was
			gReqList = req->next; // TODO got a crash here when launching a game (probabyl due to shutdown()
			free(req);
						
			if (!gReqList)
				gReqEnd = NULL;

			SignalSema(gProcSemaId);
			SignalSema(gEndSemaId);
		}
	}
	
	// delete the semaphores
	DeleteSema(gProcSemaId);
	DeleteSema(gEndSemaId);
	ExitDeleteThread();
}

static void ioSimpleActionHandler(void *data) {
	io_simpleaction_t action = (io_simpleaction_t)data;
	
	if (action)
		action();
}

void ioInit(void) {
	gIOTerminate = 0;
	gHandlerCount = 0;
	gReqList = NULL;
	gReqEnd = NULL;
	
	gDispatcherThreadID = 0;
	gThreadId = 0;
	
	gQueueSema.init_count = 1;
	gQueueSema.max_count = 1;
	gQueueSema.option = 0;

	gProcSemaId = CreateSema(&gQueueSema);
	gEndSemaId = CreateSema(&gQueueSema);
	gIOPrintfSemaId = CreateSema(&gQueueSema);
	
	ChangeThreadPriority (  GetThreadId (), 29 );

	// default custom simple action handler
	ioRegisterHandler(IO_CUSTOM_SIMPLEACTION, &ioSimpleActionHandler);
	
	memset(&gThread, 0, sizeof(gThread));
	
	gThread.func         = (void *)ioThreadDispatcher;
	gThread.stack        = disp_stack;
	gThread.stack_size   = THREAD_STACK_SIZE;
	gThread.gp_reg       = &_gp;
	gThread.initial_priority = 0;
	
	SetAlarm( 625, &ioAlarmFunc, NULL); 
	
	// this one manages the thread switching in alarm intervals
	gDispatcherThreadID = CreateThread(&gThread);
	StartThread(gDispatcherThreadID, NULL);

	gThread.func         = (void *)ioWorkerThread;
	gThread.stack        = thread_stack;
	gThread.initial_priority = 30;

	gThreadId = CreateThread(&gThread);
	StartThread(gThreadId, NULL);
	
	ChangeThreadPriority (  GetThreadId (), 30 );
}

int ioPutRequest(int type, void* data) {
	// check the type before queueing
	if (!ioGetHandler(type))
		return IO_ERR_INVALID_HANDLER;
	
	WaitSema(gEndSemaId);
	
	// We don't have to lock the tip of the queue...
	// If it exists, it won't be touched, if it does not exist, it is not being processed
	struct io_request_t* req = gReqEnd;
	
	if (!req) {
		gReqList = (struct io_request_t*)malloc(sizeof(struct io_request_t));
		req = gReqList;
		gReqEnd = gReqList;
	} else {
		req->next = (struct io_request_t*)malloc(sizeof(struct io_request_t));
		req = req->next;
		gReqEnd = req;
	}
	
	req->next = NULL;
	req->type = type;
	req->data = data;
	
	SignalSema(gEndSemaId);
	
	WakeupThread(gThreadId);
	return IO_OK;
}

int ioRemoveRequests(int type) {
	// lock the deletion sema and the queue end sema as well
	WaitSema(gEndSemaId);
	WaitSema(gProcSemaId);
	
	int count = 0;
	struct io_request_t* req = gReqList;
	struct io_request_t* last = NULL;
	
	while (req) {
		if (req->type == type) {
			struct io_request_t* next = req->next;
			
			if (last)
				last->next = next;
			
			if (req == gReqList)
				gReqList = next;
			
			if (req == gReqEnd)
				gReqEnd = last;
			
			count++;
			free(req);
			
			req = next;
		} else {
			last = req;
			req = req->next;
		}
	}
	
	SignalSema(gProcSemaId);
	SignalSema(gEndSemaId);
	
	return count;
}

void ioEnd(void) {
	// termination requested flag
	gIOTerminate = 1;
	
	WaitSema(gProcSemaId);
	
	while (gReqList) {
		struct io_request_t* req = gReqList;
		gReqList = gReqList->next;
		free(req);
	}
	SignalSema(gProcSemaId);
	
	// wake up and wait for end
	WakeupThread(gThreadId);
}

int ioGetPendingRequestCount(void) {
	int count = 0;
	
	struct io_request_t* req = gReqList;
	
	while (req) {
		count++; req = req->next; 
	}
	
	return count;
}


// From PS2DEV.ORG - Thread dispatcher
static void ioThreadDispatcher ( void* apParam ) {
	while ( 1 ) { 
		SleepThread (); 
	}  /* end while */ 
}  /* end dispatcher */ 

static void ioAlarmFunc(s32 id, u16 time, void *arg) {
	iWakeupThread ( gDispatcherThreadID ); 
	iRotateThreadReadyQueue ( 30 ); 
	iSetAlarm ( 625, &ioAlarmFunc, NULL ); 
} 


int ioPrintf(const char* format, ...) {
	WaitSema(gIOPrintfSemaId);
	
	va_list args;
	va_start(args, format);
	int ret = vprintf(format, args);
	va_end(args);
	
	SignalSema(gIOPrintfSemaId);
	return ret;
}
